#include "CommandList.hpp"

#include <Graphics/GraphicsCommon.hpp>
#include <Graphics/Device/GraphicsDevice.hpp>
#include <Graphics/Framebuffer/Framebuffer.hpp>
#include <Graphics/Pipeline/Pipeline.hpp>

#include <Graphics/Pipeline/InputAssembler/InputAssemblerDesc.hpp>
#include <Graphics/Sampler/SamplerState.hpp>

#include <Resources/BufferObject/BufferObject.hpp>
#include <Resources/BufferObject/BufferObjectDesc.hpp>

#include <Graphics/Shader/Shader.hpp>
#include <Graphics/Shader/ShaderDesc.hpp>

CommandList::CommandList(GraphicsDevice* device)
{
    GraphicsDevice_ = device;

#ifdef ENGINE_DEBUG
    DX_PRINT("Create DeferredContext", GraphicsDevice_->GetDevice()->CreateDeferredContext(0, &DeferredContext_));
#else
    GraphicsDevice_->GetDevice()->CreateDeferredContext(0, &DeferredContext_);
#endif
}

void CommandList::DrawIndexed(uint32_t indexCount, uint32_t startIndexLocation, int32_t baseVertexLocation)
{
    GraphicsDevice_->GetImmediateContext()->DrawIndexed(indexCount, startIndexLocation, baseVertexLocation);
}

void CommandList::ExecuteCommandList()
{
/*#ifdef ENGINE_DEBUG
    DX_PRINT("Finish CommandList", DeferredContext_->FinishCommandList(false, CommandList_.GetAddressOf()));
#else*/
    //DeferredContext_->FinishCommandList(false, CommandList_.GetAddressOf());
/*#endif*/
    //GraphicsDevice_->GetImmediateContext()->ExecuteCommandList(CommandList_.Get(), false);
}

void CommandList::ClearBuffer(XMFLOAT4 color)
{
    GraphicsDevice_->GetImmediateContext()->ClearRenderTargetView(FramebufferRef_->GetColorAttachmentRTV().Get(), reinterpret_cast<const float*>(&color));
    GraphicsDevice_->GetImmediateContext()->ClearDepthStencilView(FramebufferRef_->GetDepthAttachmentDSV().Get(), D3D11_CLEAR_DEPTH, 1.0f, 0);
}

DeviceObjectType CommandList::GetDeviceObjectType()
{
    return DeviceObjectType::CommandList;
}

void CommandList::BindPipeline(Pipeline* pipeline)
{
    PipelineRef_ = pipeline;

    GraphicsDevice_->GetImmediateContext()->IASetInputLayout(pipeline->GetInputLayout().Get());
    const float factor[4] = {0,0,0,0};
    GraphicsDevice_->GetImmediateContext()->OMSetBlendState(pipeline->GetBlendState().Get(), factor, D3D11_APPEND_ALIGNED_ELEMENT);
    GraphicsDevice_->GetImmediateContext()->OMSetDepthStencilState(pipeline->GetDepthStencilState().Get(), 0);
    GraphicsDevice_->GetImmediateContext()->RSSetState(pipeline->GetRasterizerState().Get());

    switch(pipeline->GetDesc().InputAssemblerDesc_->MeshTopology_) {
        case MeshTopology::TriangleList:
            GraphicsDevice_->GetImmediateContext()->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            break;
        case MeshTopology::LineList:
            GraphicsDevice_->GetImmediateContext()->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_LINELIST);
            break;
        case MeshTopology::PointList:
            GraphicsDevice_->GetImmediateContext()->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_POINTLIST);
            break;
        default:
            break;
    }

    GraphicsDevice_->GetImmediateContext()->VSSetShader(pipeline->GetVertexShader().Get(), nullptr, 0);
    GraphicsDevice_->GetImmediateContext()->PSSetShader(pipeline->GetPixelShader().Get(), nullptr, 0);
}

void CommandList::BindFramebuffer(Framebuffer* framebuffer)
{
    FramebufferRef_ = framebuffer;
    // This may take array later. So there may be a for loop.
    GraphicsDevice_->GetImmediateContext()->OMSetRenderTargets(1, framebuffer->GetColorAttachmentRTV().GetAddressOf(), nullptr);
}

void CommandList::BindViewport(XMINT2 windowSize)
{
    D3D11_VIEWPORT viewport{};

    viewport.TopLeftX = 0.f;
    viewport.TopLeftY = 0.f;
    viewport.Width = static_cast<float>(windowSize.x);
    viewport.Height = static_cast<float>(windowSize.y);
    viewport.MinDepth = 0.f;
    viewport.MaxDepth = 1.f;

    //This may need for loop in order to use with multiple viewports.
    GraphicsDevice_->GetImmediateContext()->RSSetViewports(1, &viewport);
}

void CommandList::BindSamplers(SamplerState* samplerState)
{
    GraphicsDevice_->GetImmediateContext()->PSSetSamplers(0, 1, samplerState->GetSamplerState().GetAddressOf()); // This code may need for loop in order to control multiple samplers.
}

void CommandList::BindVertexBuffer(BufferObject* vertexBuffer)
{
    uint32_t stride = vertexBuffer->GetDesc().StructureByteStride;
    uint32_t offset = 0;

    GraphicsDevice_->GetImmediateContext()->IASetVertexBuffers(0, 1, vertexBuffer->GetBuffer().GetAddressOf(), &stride, &offset);
}

void CommandList::BindIndexBuffer(BufferObject* indexBuffer)
{
    GraphicsDevice_->GetImmediateContext()->IASetIndexBuffer(indexBuffer->GetBuffer().Get(), DXGI_FORMAT_R16_UINT, 0);
}

void CommandList::BindShaders()
{
    for(auto& shader : PipelineRef_->GetDesc().Shaders_) {
        switch(shader->GetShaderType()) {
            case ShaderType::Vertex:
                GraphicsDevice_->GetImmediateContext()->VSSetShader(shader->GetVertexShader().Get(), nullptr, 0);
                break;
            case ShaderType::Pixel:
                GraphicsDevice_->GetImmediateContext()->PSSetShader(shader->GetPixelShader().Get(), nullptr, 0);
                break;
            default:
                break;
        }
    }
}
