#include "Pipeline.hpp"

#include <Graphics/GraphicsCommon.hpp>

#include <Graphics/Device/GraphicsDevice.hpp>

#include <Graphics/Shader/Shader.hpp>

#include <Graphics/Pipeline/InputAssembler/InputAssemblerDesc.hpp>
#include <Graphics/Pipeline/DepthStencil/DepthStencilStateDesc.hpp>
#include <Graphics/Pipeline/Rasterizer/RasterizerStateDesc.hpp>
#include <Graphics/Pipeline/Blend/BlendStateDesc.hpp>


DeviceObjectType Pipeline::GetDeviceObjectType()
{
    return DeviceObjectType::Pipeline;
}

Pipeline::Pipeline(PipelineDesc& desc, GraphicsDevice& device)
{
    GraphicsDevice_ = &device;
    Desc_ = desc;
}

void Pipeline::InitializePipeline()
{
    std::vector<D3D11_INPUT_ELEMENT_DESC> InputElementDesc;

    for(int i = 0; i < Desc_.InputAssemblerDesc_->SemanticName_.size(); i++) {
        D3D11_INPUT_ELEMENT_DESC desc{};

        switch(Desc_.InputAssemblerDesc_->SemanticName_[i]) {
            case InputLayoutSemanticName::Position:
                desc.SemanticName = "POSITION";
                desc.Format = DXGI_FORMAT_R32G32B32_FLOAT;
                break;
            case InputLayoutSemanticName::Normal:
                desc.SemanticName = "NORMAL";
                desc.Format = DXGI_FORMAT_R32G32B32_UINT;
                break;
            case InputLayoutSemanticName::Tangent:
                desc.SemanticName = "TANGENT";
                desc.Format = DXGI_FORMAT_R32G32B32_FLOAT;
                break;
            case InputLayoutSemanticName::Binormal:
                desc.SemanticName = "BINORMAL";
                desc.Format = DXGI_FORMAT_R32G32B32_FLOAT;
                break;
            case InputLayoutSemanticName::TexCoord:
                desc.SemanticName = "TEXCOORD";
                desc.Format = DXGI_FORMAT_R32G32_FLOAT;
                break;
            case InputLayoutSemanticName::Color:
                desc.SemanticName = "COLOR";
                desc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
                break;
            default:
                break;
        }

        desc.SemanticIndex = Desc_.InputAssemblerDesc_->SemanticIndex_[i];
        desc.InputSlot = Desc_.InputAssemblerDesc_->InputSlot_[i];
        desc.AlignedByteOffset = Desc_.InputAssemblerDesc_->AlignedByteOffset_[i];
        desc.InputSlotClass = static_cast<D3D11_INPUT_CLASSIFICATION>(Desc_.InputAssemblerDesc_->InputSlotClass_[i]);
        desc.InstanceDataStepRate = Desc_.InputAssemblerDesc_->InstanceDataStepRate_[i];

        InputElementDesc.push_back(desc);
    }

#ifdef ENGINE_DEBUG
    for(int i = 0; i < Desc_.Shaders_.size(); i++)
        if(Desc_.Shaders_[i]->GetShaderType() == ShaderType::Vertex)
            DX_PRINT("Create InputLayout", GraphicsDevice_->GetDevice()->CreateInputLayout(InputElementDesc.data(), InputElementDesc.size(),
                                                                                           Desc_.Shaders_[i]->GetShaderBlob()->GetBufferPointer(), Desc_.Shaders_[i]->GetShaderBlob()->GetBufferSize(),
                                                                                           &InputLayout_));
#else
    for(int i = 0; i < Desc_.Shaders_.size(); i++)
        if(Desc_.Shaders_[i]->GetShaderType() == ShaderType::Vertex)
            GraphicsDevice_->GetDevice()->CreateInputLayout(InputElementDesc.data(), InputElementDesc.size(),
                                                            Desc_.Shaders_[i]->GetShaderBlob()->GetBufferPointer(), Desc_.Shaders_[i]->GetShaderBlob()->GetBufferSize(),
                                                            &InputLayout_);
#endif
    mVertexShader = Desc_.Shaders_[0]->GetVertexShader();
    mPixelShader = Desc_.Shaders_[1]->GetPixelShader();

    CreateDepthStencilState();
    CreateRasterizerState();
    CreateBlendState();

}

void Pipeline::CreateDepthStencilState()
{
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc{
            .DepthEnable = Desc_.DepthStencilStateDesc_->DepthEnable,
            .DepthWriteMask = static_cast<D3D11_DEPTH_WRITE_MASK>(Desc_.DepthStencilStateDesc_->DepthWriteMask),
            .DepthFunc = static_cast<D3D11_COMPARISON_FUNC>(Desc_.DepthStencilStateDesc_->DepthFunc),
            .StencilEnable = Desc_.DepthStencilStateDesc_->StencilEnable,
            .StencilReadMask = Desc_.DepthStencilStateDesc_->StencilReadMask,
            .StencilWriteMask = Desc_.DepthStencilStateDesc_->StencilWriteMask,
    };

#ifdef ENGINE_DEBUG
    DX_PRINT("Create DepthStencilState", GraphicsDevice_->GetDevice()->CreateDepthStencilState(&depthStencilDesc, DepthState_.GetAddressOf()));
#else
    GraphicsDevice_->GetDevice()->CreateDepthStencilState(&depthStencilDesc, &DepthState_);
#endif
}

void Pipeline::CreateRasterizerState()
{
    D3D11_RASTERIZER_DESC rasterizerDesc{
            .FillMode = static_cast<D3D11_FILL_MODE>(Desc_.RasterizerStateDesc_->FillMode_),
            .CullMode = static_cast<D3D11_CULL_MODE>(Desc_.RasterizerStateDesc_->CullMode_),
            .FrontCounterClockwise = static_cast<bool>(Desc_.RasterizerStateDesc_->FrontFace_),
            .DepthBias = Desc_.RasterizerStateDesc_->DepthBias_,
            .DepthBiasClamp = Desc_.RasterizerStateDesc_->DepthBiasClamp_,
            .DepthClipEnable = Desc_.RasterizerStateDesc_->DepthClipEnable_,
            .AntialiasedLineEnable = Desc_.RasterizerStateDesc_->AntialiasedLineEnable_
    };

#ifdef ENGINE_DEBUG
    if (FAILED(GraphicsDevice_->GetDevice()->CreateRasterizerState(&rasterizerDesc, RasterizerState_.GetAddressOf())))
    {
        printf("YOO");
    }
#else
    GraphicsDevice_->GetDevice()->CreateRasterizerState(&rasterizerDesc, &RasterizerState_);
#endif


}

void Pipeline::CreateBlendState()
{
    D3D11_RENDER_TARGET_BLEND_DESC rtvBlendDesc{
            .BlendEnable = Desc_.BlendStateDesc_->RenderTargetDesc.BlendEnable,
            .SrcBlend = static_cast<D3D11_BLEND>(Desc_.BlendStateDesc_->RenderTargetDesc.SrcBlend),
            .DestBlend = static_cast<D3D11_BLEND>(Desc_.BlendStateDesc_->RenderTargetDesc.DestBlend),
            .BlendOp = static_cast<D3D11_BLEND_OP>(Desc_.BlendStateDesc_->RenderTargetDesc.BlendOp),
            .SrcBlendAlpha = static_cast<D3D11_BLEND>(Desc_.BlendStateDesc_->RenderTargetDesc.SrcBlendAlpha),
            .DestBlendAlpha = static_cast<D3D11_BLEND>(Desc_.BlendStateDesc_->RenderTargetDesc.DestBlendAlpha),
            .BlendOpAlpha = static_cast<D3D11_BLEND_OP>(Desc_.BlendStateDesc_->RenderTargetDesc.BlendOpAlpha),
            .RenderTargetWriteMask = static_cast<uint8_t>(Desc_.BlendStateDesc_->RenderTargetDesc.RenderTargetWriteMask)
    };

    D3D11_BLEND_DESC blendDesc{
            .AlphaToCoverageEnable = Desc_.BlendStateDesc_->AlphaToCoverageEnable,
            .IndependentBlendEnable = Desc_.BlendStateDesc_->IndependentBlendEnable,
            .RenderTarget = {rtvBlendDesc}
    };

#ifdef ENGINE_DEBUG
    if (FAILED(GraphicsDevice_->GetDevice()->CreateBlendState(&blendDesc, BlendState_.GetAddressOf())))
    {
        printf("yooo");
    }
#else
    GraphicsDevice_->GetDevice()->CreateBlendState(&blendDesc, &BlendState_);
#endif


}
