#include "Shader.hpp"

#include <Graphics/Device/GraphicsDevice.hpp>
#include <FileReader/FileReader.hpp>
#include <Graphics/GraphicsCommon.hpp>

Shader::Shader(ShaderDesc* desc, GraphicsDevice* graphicsDevice)
{
    GraphicsDevice_ = graphicsDevice;
    Desc_ = desc;
}

Shader::~Shader()
{
}

DeviceObjectType Shader::GetDeviceObjectType()
{
    return DeviceObjectType::Shader;
}

void Shader::CreateShader()
{
    ComPtr<ID3DBlob> errorBlob;

    std::string shaderBlob;
    std::string shaderModel;

    switch(Desc_->Type_) {
        case ShaderType::Vertex:
            shaderBlob = FileReader::GetShaderData("VertexShader");
            shaderModel = "vs_5_0";
            break;
        case ShaderType::Pixel:
            shaderBlob = FileReader::GetShaderData("PixelShader");
            shaderModel = "ps_5_0";
            break;
        case ShaderType::Domain:
            shaderBlob = FileReader::GetShaderData("DomainShader");
            shaderModel = "ds_5_0";
            break;
        case ShaderType::Geometry:
            shaderBlob = FileReader::GetShaderData("GeometryShader");
            shaderModel = "gs_5_0";
            break;
        case ShaderType::Hull:
            shaderBlob = FileReader::GetShaderData("HullShader");
            shaderModel = "hs_5_0";
            break;
        case ShaderType::Compute:
            shaderBlob = FileReader::GetShaderData("ComputeShader");
            shaderModel = "cs_5_0";
            break;
        default:
            SystemLog::ConsoleLog(LogType::Error,"Shader type not supported");
            break;
    }

    DX_PRINT("Compile Shader", D3DCompile(shaderBlob.c_str(), shaderBlob.length(), nullptr, nullptr, nullptr, Desc_->EntryPoint_.c_str(), shaderModel.c_str(), D3DCOMPILE_ENABLE_STRICTNESS, 0, Blob_.GetAddressOf(), &errorBlob));

    if(errorBlob.Get() != nullptr && errorBlob.Get()->GetBufferPointer() != nullptr) {
        SystemLog::ConsoleLog(LogType::Error, (char*) errorBlob->GetBufferPointer());
    }

    switch(Desc_->Type_) {
        case ShaderType::Vertex:
            DX_PRINT("Create VertexShader", GraphicsDevice_->GetDevice()->CreateVertexShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &VertexShader_));
            break;
        case ShaderType::Pixel:
            DX_PRINT("Create PixelShader", GraphicsDevice_->GetDevice()->CreatePixelShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &PixelShader_));
            break;
        case ShaderType::Domain:
            DX_PRINT("Create DomainShader", GraphicsDevice_->GetDevice()->CreateDomainShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &DomainShader_));
            break;
        case ShaderType::Geometry:
            DX_PRINT("Create GeometryShader", GraphicsDevice_->GetDevice()->CreateGeometryShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &GeometryShader_));
            break;
        case ShaderType::Hull:
            DX_PRINT("Create HullShader", GraphicsDevice_->GetDevice()->CreateHullShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &HullShader_));
            break;
        case ShaderType::Compute:
            DX_PRINT("Create ComputeShader", GraphicsDevice_->GetDevice()->CreateComputeShader(Blob_->GetBufferPointer(), Blob_->GetBufferSize(), nullptr, &ComputeShader_));
            break;
        default:
            SystemLog::ConsoleLog(LogType::Error,"Shader type not supported");
            break;
    }

}
